import functools
import json

from typing import Any, Callable

import websocket

from django.contrib import admin, messages
from django.db.models import QuerySet
from django.http import HttpRequest


class AdminTaskAction:
    header: dict[str, str] = {'Content-Type': 'application/json'}

    def __init__(self, task_name: str, **kwargs):
        self.task_name = task_name
        self.kwargs = kwargs

    def __call__(self, post_schedule_callable: Callable[[Any, HttpRequest, QuerySet], Any]):
        @admin.action(**self.kwargs)
        @functools.wraps(post_schedule_callable)
        def action_callable(modeladmin: admin.ModelAdmin, request: HttpRequest, queryset):
            ws_response = self.websocket_task_schedule(
                request, self.task_name, instance_ids=list(queryset.values_list('pk', flat=True))
            )
            modeladmin.message_user(request, ws_response, messages.INFO)
            return post_schedule_callable(modeladmin, request, queryset)

        return action_callable

    def websocket_task_schedule(self, http_request: HttpRequest, task_name: str, **inputs):
        ws = websocket.WebSocket()
        ws.connect(f'ws://{http_request.get_host()}/tasks/', header=self.header)
        ws.send(json.dumps([dict(name=task_name, inputs=inputs)], indent=4))
        ws_response = ws.recv()
        ws.close()
        return ws_response



import configparser
import setuptools


ini = configparser.ConfigParser()
ini.read('version.ini')


with open('README.md') as readme:
    long_description = readme.read()


tests_require = ['pytest', 'pytest-cov', 'pytest-django', 'pytest-asyncio', 'beautifulsoup4',
                 'bdd-coder==2.2.3.dev2']

setuptools.setup(
    name=ini['version']['name'],
    version=ini['version']['value'],
    author='Daniel Farré Manzorro',
    author_email='d.farre.m@gmail.com',
    description='Running background tasks trhough REST API and websocket, with channels-redis',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://bitbucket.org/coleopter/django-tasks',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers'],
    packages=setuptools.find_packages(),
    install_requires=[
        'Django', 'django-filter', 'django-extensions', 'django-request-logging', 'djangorestframework',
        'django-bootstrap-v5', 'channels', 'channels-redis', 'daphne', 'tzdata', 'psycopg2-binary',
        'websocket-client'],
    extras_require={'dev': ['ipdb', 'ipython'],
                    'mypy': ['mypy', 'django-stubs', 'djangorestframework-stubs[compatible-mypy]',
                             'types-beautifulsoup4'],
                    'test': tests_require},
    tests_require=tests_require
)



import collections
import inspect
import importlib

from typing import Callable, Optional, Any

from django.conf import settings


class TaskCoroInfo:
    def __init__(self, name: str, **inputs: dict[str, Any]):
        self.name = name.strip()
        self.inputs = inputs

    @property
    def callable(self) -> Optional[Callable]:
        module_names = settings.DJANGO_TASKS.get('coroutine_modules', [])

        for module_name in module_names:
            try:
                module = importlib.import_module(module_name)
            except ImportError:
                continue
            else:
                callable = getattr(module, self.name, None)

                if inspect.iscoroutinefunction(callable):
                    return callable
        else:
            return None

    @property
    def parameter_keys(self) -> tuple[set[str], set[str]]:
        callable = self.callable
        assert callable is not None

        params = inspect.signature(callable).parameters
        required_keys = set(k for k, v in params.items() if v.default == inspect._empty)
        optional_keys = set(k for k, v in params.items() if v.default != inspect._empty)

        return required_keys, optional_keys

    @property
    def task_call_errors(self) -> dict[str, list[str]]:
        errors = collections.defaultdict(list)

        if self.callable is None:
            errors['name'].append(f"Coroutine '{self.name}' not found.")
            return errors

        required_keys, optional_keys = self.parameter_keys
        input_keys = set(self.inputs)
        missing_keys = required_keys - input_keys
        unknown_keys = input_keys - required_keys - optional_keys

        if missing_keys:
            errors['inputs'].append(f'Missing required parameters {missing_keys}.')

        if unknown_keys:
            errors['inputs'].append(f'Unknown parameters {unknown_keys}.')

        return errors
